#include <sys/types.h>
#include <sys/ioctl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <locale.h>
#include <err.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "perf_util.h"

#include "jvmti.h"
#include "jni.h"


static int   perfevent_num_fds = 0;
static char* perfevent_events  = NULL;
static int*  perfevent_fds     = NULL;

static long cputime_begin;//million seconds
static long cputime_end;
static long wallclock_begin;
static long wallclock_end;

long get_cputime(void)
{
  struct rusage cputime;
  getrusage(RUSAGE_SELF,&cputime);
  return (cputime.ru_utime.tv_sec + cputime.ru_stime.tv_sec)*1000 + (cputime.ru_utime.tv_usec + cputime.ru_stime.tv_usec)/1000;
}

long get_wallclock(void)
{
  struct timeval time;
  gettimeofday(&time,NULL);
  return time.tv_sec*1000 + time.tv_usec/1000;
}

void perfevent_init() {
  char *fd_string;
  char *events;
  char *cursor;
  int i, read;

  events = getenv("PERF_EVENT_NAMES");
  if (!events)
    errx(1, "error reading environment variable PERF_EVENT_NAMES");

  fd_string = getenv("PERF_EVENT_FDS");
  if (!fd_string)
    errx(1, "error reading environment variable PERF_EVENT_FDS");

  perfevent_events = (char*)malloc(strlen(events)+1);
  if (!perfevent_events)
    errx(1, "malloc of perfevent_events failed");
  strcpy(perfevent_events, events);

  perfevent_num_fds = 1;
  cursor = perfevent_events;
  while (*cursor++)
    if (*cursor == ' ')
      perfevent_num_fds++;

  printf("JNI: %d events:%s\n"
	 "JNI: fd list:%s\n",
	 perfevent_num_fds,
	 events,
	 fd_string);

  perfevent_fds = (int*)malloc(sizeof(int)*perfevent_num_fds);
  if (!perfevent_fds)
    errx(1, "malloc of perfevent_fds failed");

  cursor=fd_string;
  for(i=0; i<perfevent_num_fds; i++) {
    read = sscanf(cursor, "%d", perfevent_fds+i);
    if (read != 1)
      errx(1, "sscanf of PERF_EVENT_FDS failed %d %s", read, cursor);
    while(*cursor++!=' ');      
  }

  cputime_begin = get_cputime();
  
  wallclock_begin = get_wallclock();
}

void perfevent_begin() {
  int ret;
  ret = ioctl(perfevent_fds[0], PERF_EVENT_IOC_ENABLE, 0);
  if (ret)
    err(1, "ioctl(enable) failed");
}
void perfevent_end() {
  int ret;
  ret = ioctl(perfevent_fds[0], PERF_EVENT_IOC_DISABLE, 0);
  if (ret)
    err(1, "ioctl(disable) failed");
}
void perfevent_reset() {
  int i, ret;
  for(i=0; i < perfevent_num_fds; i++) {
    ret = ioctl(perfevent_fds[i], PERF_EVENT_IOC_RESET, 0);
    if (ret)
      err(1, "ioctl(reset) failed");
  }
  cputime_begin = get_cputime();
  wallclock_begin = get_wallclock();
}
void perfevent_read() {
  int i, ret;
  uint64_t values[3];
  memset(values, 0, sizeof(values));

  fprintf(stderr, "============================ Tabulate Statistics ============================\n");
  fprintf(stderr, "CPUTIMES WALLCLOCK CPUUSAGE ");
  fprintf(stderr, "%s", perfevent_events);
  fprintf(stderr, "\n");

  cputime_end = get_cputime();
  wallclock_end = get_wallclock();
  fprintf(stderr,"%ld\t%ld\t%ld\t",cputime_end-cputime_begin, wallclock_end-wallclock_begin,(cputime_end-cputime_begin)*100/(wallclock_end-wallclock_begin));
  
  

  for (i=0; i < perfevent_num_fds; i++) {
    uint64_t val;
    double ratio;

    ret = read(perfevent_fds[i], values, sizeof(values));
    if (ret < sizeof(values)) {
      if (ret == -1)
        err(1, "cannot read results: %s", strerror(errno));
      else
        warnx("could not read event%d", i);
    }
    val = perf_scale(values);
    ratio = perf_scale_ratio(values);

    if (ratio == 1.0)
      fprintf(stderr, "%lld	", (long long int) val);
    else
      if (ratio == 0.0)
        fprintf(stderr, "NO_VALUE\t");
      else
        fprintf(stderr, "%lld-SCALED-%.2f%%	", (long long int) val, ratio*100.0);
  }
  fprintf(stderr, "\n=============================================================================\n");
}
void perfevent_cleanup() {
  int i;
  for (i=0; i < perfevent_num_fds; i++) {
    close(perfevent_fds[i]);
  }
  free(perfevent_fds);
  free(perfevent_events);
  perfevent_num_fds = 0;
  perfevent_events = NULL;
  perfevent_fds = NULL;
}

/* JNI Functions */
JNIEXPORT void JNICALL Java_probe_PerfEventLauncherProbe_init
  (JNIEnv *env, jobject o) {
  perfevent_init();
}
JNIEXPORT void JNICALL Java_probe_PerfEventLauncherProbe_begin
  (JNIEnv *env, jobject o, jstring benchmark, jint iteration, jboolean warmup) {
  //  perfevent_begin();
  perfevent_reset();
}
JNIEXPORT void JNICALL Java_probe_PerfEventLauncherProbe_end
  (JNIEnv *env, jobject o, jstring benchmark, jint iteration, jboolean warmup) {
  //  perfevent_end();
  perfevent_read();
}
JNIEXPORT void JNICALL Java_probe_PerfEventLauncherProbe_report
  (JNIEnv *env, jobject o, jstring benchmark, jint iteration, jboolean warmup) {
  //  perfevent_read();
  //  perfevent_reset();
}
JNIEXPORT void JNICALL Java_probe_PerfEventLauncherProbe_cleanup
  (JNIEnv *env, jobject o) {
  perfevent_cleanup();
}
