/*
 * Kernel module which does the calculation of the criticality stacks
 *
 */

#include <linux/module.h>    
#include <linux/kernel.h> 
#include <linux/ktime.h>   
#include <linux/kprobes.h>
#include <linux/limits.h>
#include <linux/init.h>       
#include <linux/slab.h> 
#include <linux/spinlock.h> 
#include "mod_calc_stacks.h"

#define DEBUG 0 // 1 means activate logging with printk
#define MAX_THREADS 200 // max number of threads in stack

#define LOG_FILTER "[CRITICALITY_STACKS] " 

#define ITERATION_TO_PROFILE 		"13" 

long long lastTimeStamp;
int numCriticalThreads;
long long otherCritical;
int numThreads;
spinlock_t my_lock;
int numAppThreads;
int numCalledExit;

long long concGCCritical;
long long concGCCriticalStart;

bool started;
bool counting;

typedef struct {
	pid_t pid; // pid of thread
	bool active; // indicates if thread is active
	bool futexWait; // indicates if thread is waiting on futex
	bool exited; // indicates thread already called do_exit
	long double criticality; // criticality count for thread
	long long runtime;
	long long nrFutexWaits; // number of times thread called futex.wait

	bool ignore;

	long long futexWaiting;
	long long exitWaiting;
	long long otherWaiting;
} thread_struct;

thread_struct* threads; // list of thread structs

char process_name[NAME_MAX];
char iteration_number[NAME_MAX];


inline void countWaitingTime(thread_struct *thread, long long ts, long long lastTimestamp) {
	if (thread->futexWait) {
		thread->futexWaiting += (ts - lastTimeStamp);
	} else if (thread->exited && numCalledExit < numAppThreads) {
		thread->exitWaiting += (ts - lastTimeStamp);
	} else {
		thread->otherWaiting += (ts - lastTimeStamp);
	}
}


/* signal start of the program */
void signalProgramStart(char* p_name, long long ts) {

	if (!started) {
		lastTimeStamp = ts;
		numCriticalThreads = 1;
		numThreads = 1;
		
#if DEBUG
			printk(KERN_INFO "process %s started at time: %lld\n", p_name, ts);
#endif

		started = true;
	} else {
		printk(KERN_INFO "DON'T START SAME PROGRAM TWICE WITHOUT EXITING!\n");
	}
}

/* signal start of new thread */
void signalThreadStart(pid_t pid, char* p_name, long long ts) {	

	spin_lock(&my_lock);

	++numThreads;

#if DEBUG
		printk(KERN_INFO "new thread started at time: %lld, parent_pid: %d, p_name: %s\n", ts, pid, p_name);
#endif
		
	if (numThreads > MAX_THREADS) {
		printk(KERN_INFO "Panic: more threads created than accepted!!");
	}

	spin_unlock(&my_lock);
}

/* signal end of a thread or program */
void signalThreadExit(pid_t pid, char* p_name, long long ts) {	
	int i;
	bool found;

	spin_lock(&my_lock);

	found = false;

	if (counting && numCriticalThreads == 0) {
		otherCritical += (ts - lastTimeStamp);
	}

	for ( i = 0 ; i < numThreads ; i++ ) {
		if (threads[i].pid == 0) {
			/* stop iterating the list */
			break;
		} else {
			if (counting && threads[i].active && (numCriticalThreads > 0) ) {
				threads[i].criticality += ((ts - lastTimeStamp)*1.0)/(numCriticalThreads*1.0);
				threads[i].runtime += (ts - lastTimeStamp);
			}			
	
			if (counting && !threads[i].active) {
				countWaitingTime(&threads[i], ts, lastTimeStamp);
			}
	
			if (threads[i].pid == pid) {
				threads[i].active = false;
				threads[i].futexWait = false;
				threads[i].exited = true;
				found = true;

				numCalledExit++;
			}
		}
	}		

	if ( !found ) { 
		
		//printk("Added new thread with pid:%d, in thread_exit method\n",pid);
	
		threads[i].pid = pid;
		threads[i].active = false;
		threads[i].futexWait = false;
		threads[i].exited = true;
		
		if (counting && numCriticalThreads > 0) {
			threads[i].criticality += ((ts - lastTimeStamp)*1.0)/(numCriticalThreads*1.0);
			threads[i].runtime += (ts - lastTimeStamp);
		}
	}

	lastTimeStamp = ts;
	if (numCriticalThreads > 0) {
		--numCriticalThreads;
	}

#if DEBUG
		printk(KERN_INFO "thread called do_exit at time: %lld, pid: %d, p_name: %s\n", ts, pid, p_name);
#endif

	spin_unlock(&my_lock);
}

/* signal futex wait */
void signalFutexWait(pid_t pid, char* p_name, long long ts) {	
	int i;
	bool found;

	spin_lock(&my_lock);

	found = false;

	if (counting && numCriticalThreads == 0) {
		/* this should not happen */
		otherCritical += (ts - lastTimeStamp);
	}

	for ( i = 0 ; i < numThreads ; i++ ) {
		if (threads[i].pid == 0) {
			/* stop iterating the list */
			break;
		} else {
			if (counting && threads[i].active && (numCriticalThreads > 0) ) {
				threads[i].criticality += ((ts - lastTimeStamp)*1.0)/(numCriticalThreads*1.0);
				threads[i].runtime += (ts - lastTimeStamp);
			}	
			
			if (counting && !threads[i].active) {
				countWaitingTime(&threads[i], ts, lastTimeStamp);
			}		
	
			if (threads[i].pid == pid) {
				threads[i].active = false;
				threads[i].futexWait = true;
				threads[i].nrFutexWaits += 1;
				found = true;
			}
		}
	}

	if ( !found ) { 
	
		//printk("Added new thread with pid:%d, in futex wait method\n",pid);
	
		threads[i].pid = pid;
		threads[i].nrFutexWaits = 1;
		threads[i].active = false;
		threads[i].futexWait = true;
		threads[i].exited = false;
		
		if (counting && numCriticalThreads > 0) {
			threads[i].criticality += ((ts - lastTimeStamp)*1.0)/(numCriticalThreads*1.0);
			threads[i].runtime += (ts - lastTimeStamp);
		}
	}

		
	lastTimeStamp = ts;
	if (numCriticalThreads > 0) {
		--numCriticalThreads;
	}

#if DEBUG
		printk(KERN_INFO "futex_wait called at time: %lld, pid: %d, p_name: %s\n", ts, pid, p_name);
#endif

	spin_unlock(&my_lock);
}

/* signal return of futex wait function */
void signalFutexEnds(pid_t pid, char* p_name, long long ts) {
	int i;
	bool found;

	spin_lock(&my_lock);

	found = false;

	if (counting && numCriticalThreads == 0) {
		/* this should not happen */
		otherCritical += (ts - lastTimeStamp);
	}

	for ( i = 0 ; i < numThreads ; i++ ) {
		if (threads[i].pid == 0) {
			/* stop iterating the list */
			break;
		} else {
			if (counting && threads[i].active && (numCriticalThreads > 0) ) {
				threads[i].criticality += ((ts - lastTimeStamp)*1.0)/(numCriticalThreads*1.0);
				threads[i].runtime += (ts - lastTimeStamp);
			}	
			
			if (counting && !threads[i].active) {
				countWaitingTime(&threads[i], ts, lastTimeStamp);
			}		
	
			if (threads[i].pid == pid) {
				threads[i].active = true;
				threads[i].futexWait = false;
				found = true;
			}
		}
	}

	if ( !found ) { 
		/* should not happen */
		threads[i].pid = pid;
		threads[i].active = true;
		threads[i].futexWait = false;
		threads[i].exited = false;
		
		//printk("Added new thread with pid:%d in futex ends method\n",pid);
	}

	lastTimeStamp = ts;
	if (numCriticalThreads < numThreads) {
		++numCriticalThreads;
	}
	

#if DEBUG
		printk(KERN_INFO "futex_wait returned at time: %lld, pid: %d, p_name: %s\n", ts, pid, p_name);
#endif

	spin_unlock(&my_lock);
}

/* signal a thread is being scheduled out */
void signalThreadScheduledOut(pid_t pid, char* p_name, long long ts) {	
	int i;
	bool processingFutex, found, calledExit;

	spin_lock(&my_lock);

	processingFutex = false;
	found = false;
	calledExit = false;

	for ( i = 0 ; i < numThreads ; i++ )  {
		if (threads[i].pid == 0) {
			break;
		}

		if (threads[i].pid == pid && threads[i].futexWait) {
			processingFutex = true;
			break;
		}
	}

	/* if the thread is being scheduled out while waiting for a futex, this event can be neglected */
	if (!processingFutex) {
		if (counting && numCriticalThreads == 0) {
			otherCritical += (ts - lastTimeStamp);
		}

		for ( i = 0 ; i < numThreads ; i++ ) {
			if (threads[i].pid == 0) {
				/* stop iterating the list */
				break;
			} else {
				if (counting && threads[i].active && (numCriticalThreads > 0) && !threads[i].exited ) {
					threads[i].criticality += ((ts - lastTimeStamp)*1.0)/(numCriticalThreads*1.0);
					threads[i].runtime += (ts - lastTimeStamp);
				}			
	
				if (counting && !threads[i].active) {
					countWaitingTime(&threads[i], ts, lastTimeStamp);
				}
	
				if (threads[i].pid == pid) {
					threads[i].active = false;
					threads[i].futexWait = false;

					if (threads[i].exited) {
						calledExit = true;
					}					

					found = true;
				}
			}
		}

		if ( !found ) { 
			//printk("Added new thread with pid:%d in scheduled out method\n",pid);
		
			threads[i].pid = pid;
			threads[i].active = false;
			threads[i].futexWait = false;
			threads[i].exited = false;

			if (counting && numCriticalThreads > 0) {
				threads[i].criticality += ((ts - lastTimeStamp)*1.0)/(numCriticalThreads*1.0);
				threads[i].runtime += (ts - lastTimeStamp);
			}
		}

		lastTimeStamp = ts;
		if (!calledExit && numCriticalThreads > 0) {
			--numCriticalThreads;
		}
	}

#if DEBUG
		printk(KERN_INFO "thread scheduled out at time: %lld, pid: %d, p_name: %s\n", ts, pid, p_name);
#endif

	spin_unlock(&my_lock);
}

/* signal a thread is being scheduled in */
void signalThreadScheduledIn(pid_t pid, char* p_name, long long ts) {	
	int i;
	bool processingFutex, found, calledExit;

	spin_lock(&my_lock);

	processingFutex = false;
	found = false;
	calledExit = false;

	for ( i = 0 ; i < numThreads ; i++ ) {
		if (threads[i].pid == 0) {
			break;
		}

		if (threads[i].pid == pid && threads[i].futexWait) {
			processingFutex = true;
			break;
		}
	}

	/* if the thread is being scheduled out while waiting for a futex, this event can be neglected */
	if (!processingFutex) {
		if (counting && numCriticalThreads == 0) {
			otherCritical += (ts - lastTimeStamp);
		}

		for ( i = 0 ; i < numThreads ; i++ ) {
			if (threads[i].pid == 0) {
				/* stop iterating the list */
				break;
			} else {
				if (counting && threads[i].active && (numCriticalThreads > 0) && !threads[i].exited ) {
					threads[i].criticality += ((ts - lastTimeStamp)*1.0)/(numCriticalThreads*1.0);
					threads[i].runtime += (ts - lastTimeStamp);
				}	
				
				if (counting && !threads[i].active) {
					countWaitingTime(&threads[i], ts, lastTimeStamp);
				}		
	
				if (threads[i].pid == pid) {
					threads[i].active = true;
					threads[i].futexWait = false;
					found = true;

					if (threads[i].exited) {
						calledExit = true;
					}
				}
			}
		}

		if ( !found ) { 
		
			//printk("Added new thread with pid:%d in scheduled in method\n",pid);
			
			threads[i].pid = pid;
			threads[i].active = true;
			threads[i].futexWait = false;
			threads[i].exited = false;
		}

		lastTimeStamp = ts;
		if (!calledExit && numCriticalThreads < numThreads) {
			++numCriticalThreads;
		}
	}

#if DEBUG
		printk(KERN_INFO "thread scheduled in at time: %lld, pid: %d, p_name: %s\n", ts, pid, p_name);
#endif

	spin_unlock(&my_lock);
}


void signalProgramName(char* name) {
	strlcpy(process_name, name, strlen(name));
}

void signalIterationNumber(char* number){
	strlcpy(iteration_number, number, strlen(number));
}


/* method to print out the stack */
void signalPrintStack() {
	int i; 

	spin_lock(&my_lock);

	if (strcmp(iteration_number,ITERATION_TO_PROFILE) == 0 ) {
		// print out stack at the end of the program
		printk(KERN_INFO LOG_FILTER ";;;START:%s,ITERATION:%s\n", process_name, iteration_number);

		if (threads) {
			for ( i = 0 ; i < numThreads ; i++) {
				if (!threads[i].ignore) {
					// uncomment this for printing out bottle graphs
					printk(KERN_INFO LOG_FILTER ";;;%d,%lld,%lld,%lld,%lld,%lld\n", threads[i].pid, (long long) threads[i].criticality, threads[i].runtime, (long long) threads[i].futexWaiting, (long long) threads[i].exitWaiting, (long long) threads[i].otherWaiting);
				}
				// uncomment this for printing out number of futex wait events	
				//printk(KERN_INFO LOG_FILTER ";;;%d,%lld\n", threads[i].pid, threads[i].nrFutexWaits);
			}
		}

		printk(KERN_INFO LOG_FILTER ";;;concGCCritical,%lld\n", concGCCritical);
		printk(KERN_INFO LOG_FILTER ";;;END\n");
	}

	spin_unlock(&my_lock);
}

void signalResetStack() {
	int i;

	spin_lock(&my_lock);

	for ( i = 0 ; i < MAX_THREADS ; i++) {
		if (threads[i].pid == 0) {
			break;
		}

		threads[i].criticality = 0;
		threads[i].runtime = 0;
		threads[i].nrFutexWaits = 0;
		threads[i].futexWaiting = 0;
		threads[i].exitWaiting = 0;
		threads[i].otherWaiting = 0;

		if (threads[i].exited) {	
			threads[i].ignore = true;
			/*
			threads[i].pid = 0;
			--numThreads;
			*/
		}

	
	}

	otherCritical = 0;
	numCalledExit = 0;
	concGCCritical = 0;
	concGCCriticalStart = -1;

	spin_unlock(&my_lock);
}

void signalClearStack() {
	spin_lock(&my_lock);

	if (threads){
		kfree(threads);
		threads = NULL;
	}
		
	otherCritical = 0;

	spin_unlock(&my_lock);
}

void signalInitStack() {
	int i;

	spin_lock(&my_lock);

	/* initialization of variables */
	lastTimeStamp = 0;
	otherCritical = 0;
	numCriticalThreads = 0;
	numThreads = 0;
	numCalledExit = 0;
	concGCCritical = 0;
	concGCCriticalStart = -1;

	/* allocate and initialize memory */
	threads = kmalloc ( MAX_THREADS * sizeof(thread_struct) , GFP_KERNEL );
	
	for ( i = 0 ; i < MAX_THREADS ; i++ ) {
		threads[i].pid = 0;
		threads[i].active = false;
		threads[i].futexWait = false;
		threads[i].exited = false;
		threads[i].criticality = 0;
		threads[i].runtime = 0;
		threads[i].nrFutexWaits = 0;
		threads[i].futexWaiting = 0;
		threads[i].exitWaiting = 0;
		threads[i].otherWaiting = 0;
		threads[i].ignore = false;
	}

	started = false;
	counting = false;

	spin_unlock(&my_lock);
}

/* signal start of collecting data for stacks */
void signalStartCollecting() {
	//spin_lock(&my_lock);

	counting = true;
	lastTimeStamp = (long long) ktime_to_ns(ktime_get());

	signalResetStack();	

	//spin_unlock(&my_lock);
}


/* signal to stop collecting data for stacks */
void signalStopCollecting() {
	long long ts;
	int i;

	spin_lock(&my_lock);

	ts = (long long) ktime_to_ns(ktime_get());

	for ( i = 0 ; i < numThreads ; i++ ) {
		if (threads[i].pid == 0) {
			/* stop iterating the list */
			break;
		} else {
			if (counting && threads[i].active && (numCriticalThreads > 0) && !threads[i].exited ) {
				threads[i].criticality += ((ts - lastTimeStamp)*1.0)/(numCriticalThreads*1.0);
				threads[i].runtime += (ts - lastTimeStamp);
			}			
		}
	}

	counting = false;

	spin_unlock(&my_lock);
}

void signalNumAppThreads(int numThreads) {
	numAppThreads = numThreads;
}

void signalConcGCSTW(bool active) {
	long long timeStamp = (long long) ktime_to_ns(ktime_get());

	if (active) {
		concGCCriticalStart = timeStamp;
	} else {
		if ( concGCCriticalStart > 0 ) {
			concGCCritical += ( timeStamp - concGCCriticalStart );
			concGCCriticalStart = -1;
		}
	}
}

/* init function of module */
static int __init mod_calc_stacks_init(void) {	
	/* initialization of spinlock */
	spin_lock_init(&my_lock);

	signalInitStack();

	printk(KERN_INFO "Module for computing criticality stacks loaded!\n");  

	return 0; 
}

/* exit function of module */
static void __exit mod_calc_stacks_cleanup(void) {
	/*int i;
		
	// print out stack at the end of the program
	printk(KERN_INFO "Stack at the end of the program:\n");

	for ( i = 0 ; i < numThreads ; i++) {
		printk(KERN_INFO "\t\tThread %d => %lld\n", threads[i].pid, threads[i].criticality);
	}

	printk(KERN_INFO "\t\tOther => %lld\n", otherCritical);*/
	
	// free allocated memory 
	if (threads){
		kfree(threads);
		threads = NULL;
	}
	
	printk(KERN_INFO "Cleaning up module for computing criticality stacks.\n");
}

/* Expose functions to other modules */
EXPORT_SYMBOL(signalProgramStart); 
EXPORT_SYMBOL(signalThreadStart);
EXPORT_SYMBOL(signalThreadExit);
EXPORT_SYMBOL(signalFutexWait);
EXPORT_SYMBOL(signalFutexEnds);
EXPORT_SYMBOL(signalThreadScheduledOut);
EXPORT_SYMBOL(signalThreadScheduledIn);

EXPORT_SYMBOL(signalPrintStack);
EXPORT_SYMBOL(signalResetStack);
EXPORT_SYMBOL(signalClearStack);
EXPORT_SYMBOL(signalInitStack);
EXPORT_SYMBOL(signalStartCollecting);
EXPORT_SYMBOL(signalStopCollecting);

EXPORT_SYMBOL(signalProgramName);
EXPORT_SYMBOL(signalIterationNumber);

EXPORT_SYMBOL(signalNumAppThreads);
EXPORT_SYMBOL(signalConcGCSTW);

module_init(mod_calc_stacks_init);
module_exit(mod_calc_stacks_cleanup);
MODULE_LICENSE("GPL");
