/*
 * Probe that tracks when a new thread is created
 *
 * usage: insmod probe_clone.ko process=<process_name>
 * parameter <process_name> is the name of the process we want to monitor,
 * default process is bash
 */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kprobes.h>
#include <linux/limits.h>
#include <linux/ktime.h>


static char process_name[NAME_MAX] = "bash";
module_param_string(process, process_name, NAME_MAX, S_IRUGO);
MODULE_PARM_DESC(process, "Process to monitor with kprobe");


/* defined in mod_calc_stacks.h, implemented in mod_calc_stacks.c */
void signalThreadStart(pid_t, char*, long long);


/* For each probe you need to allocate a kprobe structure */
static struct kprobe kp = {
	.symbol_name	= "sys_clone",
};  

/* kprobe pre_handler: called just before the probed instruction is executed */
static int handler_pre(struct kprobe *p, struct pt_regs *regs) {
	/* cur_p_name contains name current process */	
	char cur_p_name[NAME_MAX];
	strlcpy(cur_p_name, current->comm, sizeof(cur_p_name));	
	
	if ( strcmp(cur_p_name,process_name) == 0 ) {
		signalThreadStart(current->pid, cur_p_name, (long long) ktime_to_ns(ktime_get()));
		//printk(KERN_INFO "new thread started at time: %lld, parent_pid: %d, p_name: %s\n", (long long) ktime_to_ns(ktime_get()), current->pid, cur_p_name);
	}

	return 0;
}

/* kprobe post_handler: called after the probed instruction is executed */
static void handler_post(struct kprobe *p, struct pt_regs *regs, unsigned long flags) { /* do nothing */ }

/*
 * fault_handler: this is called if an exception is generated for any
 * instruction within the pre- or post-handler, or when Kprobes
 * single-steps the probed instruction.
 */
static int handler_fault(struct kprobe *p, struct pt_regs *regs, int trapnr) {
	printk(KERN_INFO "fault_handler: p->addr = 0x%p, trap #%dn", p->addr, trapnr);
	return 0;
}

static int __init kprobe_init(void) {
	int ret;

	kp.pre_handler = handler_pre;
	kp.post_handler = handler_post;
	kp.fault_handler = handler_fault;

	ret = register_kprobe(&kp);
	if (ret < 0) {
		printk(KERN_INFO "register_kprobe failed, returned %d\n", ret);
		return ret;
	}
	printk(KERN_INFO "Planted kprobe at %p\n", kp.addr);

	return 0;
}

static void __exit kprobe_exit(void) {
	unregister_kprobe(&kp);
	printk(KERN_INFO "kprobe at %p unregistered\n", kp.addr);
}

module_init(kprobe_init)
module_exit(kprobe_exit)
MODULE_LICENSE("GPL");
