/*
 * Probe that tracks when futex_wait function is executed 
 * and when the function returns (= finishes)
 *
 * usage: insmod probe_futex_wait.ko process=<process_name>
 * parameter <process_name> is the name of the process we want to monitor,
 * default process is bash
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kprobes.h>
#include <linux/ktime.h>
#include <linux/limits.h>
#include <linux/sched.h>

#include "probe_futex_wait.h"

static char func_name[NAME_MAX] = "futex_wait";
static char process_name[NAME_MAX];
//module_param_string(process, process_name, NAME_MAX, S_IRUGO);
//MODULE_PARM_DESC(process, "Process to monitor with probe");


/* defined in mod_calc_stacks.h, implemented in mod_calc_stacks.c */
void signalFutexWait(pid_t, char*, long long);
void signalFutexEnds(pid_t, char*, long long);

void signalProgramNameProbeFutexWait(char* p_name) {
	strlcpy(process_name, p_name, strlen(p_name));	
	printk (KERN_INFO "process_name in probe_futex_wait.c set to %s\n", process_name);
}

/* Here we use the entry_handler to timestamp function entry */
static int entry_handler(struct kretprobe_instance *ri, struct pt_regs *regs) {
	/* cur_p_name contains name current process */	
	//char cur_p_name[NAME_MAX];
	//strlcpy(cur_p_name, current->comm, sizeof(cur_p_name));	

	if (strlen(process_name) > 0 && strcmp(current->comm,process_name) == 0 ) {
		signalFutexWait(current->pid, current->comm, (long long) ktime_to_ns(ktime_get()));
		//printk(KERN_INFO "futex_wait called at time: %lld, pid: %d, p_name: %s\n", (long long) ktime_to_ns(ktime_get()), current->pid, cur_p_name);
	}	
	
	return 0;
}

/*
 * Return-probe handler: Log the return value and duration. Duration may turn
 * out to be zero consistently, depending upon the granularity of time
 * accounting on the platform.
 */
static int ret_handler(struct kretprobe_instance *ri, struct pt_regs *regs) {
	/* cur_p_name contains name current process */	
	//char cur_p_name[NAME_MAX];
	//strlcpy(cur_p_name, current->comm, sizeof(cur_p_name));	

	if (strlen(process_name) > 0 && strcmp(current->comm,process_name) == 0 ) {
		signalFutexEnds(current->pid, current->comm, (long long) ktime_to_ns(ktime_get()));
		//printk(KERN_INFO "futex_wait returned at time: %lld, pid: %d, p_name: %s\n", (long long) ktime_to_ns(ktime_get()), current->pid, cur_p_name);
	}

	return 0;
}

static struct kretprobe my_kretprobe = {
	.handler		= ret_handler,
	.entry_handler		= entry_handler,
	.data_size		= 0,
	/* Probe up to 512 instances concurrently. */
	.maxactive		= 512,
};

static int __init kretprobe_init(void) {
	int ret;

	my_kretprobe.kp.symbol_name = func_name;
	ret = register_kretprobe(&my_kretprobe);
	if (ret < 0) {
		printk(KERN_INFO "register_kretprobe failed, returned %d\n", ret);
		return -1;
	}
	printk(KERN_INFO "Planted return probe at %s: %p\n", my_kretprobe.kp.symbol_name, my_kretprobe.kp.addr);

	return 0;
}

static void __exit kretprobe_exit(void) {
	unregister_kretprobe(&my_kretprobe);
	printk(KERN_INFO "kretprobe at %p unregistered\n", my_kretprobe.kp.addr);

	/* nmissed > 0 suggests that maxactive was set too low. */
	printk(KERN_INFO "Missed probing %d instances of %s\n", my_kretprobe.nmissed, my_kretprobe.kp.symbol_name);
}

/* Expose functions to other modules */
EXPORT_SYMBOL(signalProgramNameProbeFutexWait); 

module_init(kretprobe_init)
module_exit(kretprobe_exit)
MODULE_LICENSE("GPL");
