#include <linux/module.h>	
#include <linux/kernel.h>	
#include <linux/proc_fs.h>	
#include <linux/limits.h>
#include <linux/spinlock.h> 
#include <asm/uaccess.h>	


spinlock_t my_lock;

static struct proc_dir_entry *proc_file_process_name;
static char process_name_buffer[NAME_MAX];
static unsigned long process_name_buffer_size = 0;

static struct proc_dir_entry *proc_file_action;
static char action_buffer[NAME_MAX];
static unsigned long action_buffer_size = 0;

static struct proc_dir_entry *proc_output_file_name;
static char output_file_name_buffer[NAME_MAX];
static unsigned long output_file_name_buffer_size = 0;

static struct proc_dir_entry *iteration_number;
static char iteration_number_buffer[NAME_MAX];
static unsigned long iteration_number_buffer_size = 0;

static struct proc_dir_entry *proc_num_app_threads;
static char num_app_threads_buffer[NAME_MAX];
static unsigned long num_app_threads_buffer_size = 0;

static struct proc_dir_entry *proc_conc_gc_stw;
static char conc_gc_stw_buffer[NAME_MAX];
static unsigned long conc_gc_stw_buffer_size = 0;


// external functions that need to be called when process_name is set
void signalProgramNameProbeExecve(char*);
void signalProgramNameProbeSchedSwitch(char*);
void signalProgramNameProbeFutexWait(char*);
void signalProgramNameProbeDoExit(char*);
void signalProgramNameProbeDoFork(char*);

// external functions used for signaling actions
void signalStartCollecting(void);
void signalStopCollecting(void);
void signalPrintStack(void);
void signalResetStack(void);
void signalClearStack(void);
void signalInitStack(void);

void signalProgramName(char*);
void signalIterationNumber(char*);

void signalNumAppThreads(int);
void signalConcGCSTW(bool);

/*
 * This function is called when the /proc file is read
 */
int procfile_read(char *buffer, char **buffer_location, off_t offset, int buffer_length, int *eof, void *data) {
	int ret;
	
	if (offset > 0) {
		// we have finished to read, return 0
		ret  = 0;
	} else {
		// fill the buffer, return the buffer size
		memcpy(buffer, process_name_buffer, process_name_buffer_size);
		ret = process_name_buffer_size;
	}

	return ret;
}

/*
 * This function is called when the /proc file is written
 */
int procfile_write(struct file *file, const char *buffer, unsigned long count, void *data) {
	// get buffer size
	process_name_buffer_size = count;
	if (process_name_buffer_size > NAME_MAX ) {
		process_name_buffer_size = NAME_MAX;
	}
	
	// write data to the buffer 
	if ( copy_from_user(process_name_buffer, buffer, process_name_buffer_size) ) {
		return -EFAULT;
	}
	
	// signal program name to probes
	signalProgramNameProbeExecve(process_name_buffer);
	signalProgramNameProbeSchedSwitch(process_name_buffer);
	signalProgramNameProbeFutexWait(process_name_buffer);
	signalProgramNameProbeDoExit(process_name_buffer);
	signalProgramNameProbeDoFork(process_name_buffer);

	return process_name_buffer_size;
}

int proc_actionfile_read(char *buffer, char **buffer_location, off_t offset, int buffer_length, int *eof, void *data) {
	int ret;
	
	if (offset > 0) {
		ret  = 0;
	} else {
		memcpy(buffer, action_buffer, action_buffer_size);
		ret = action_buffer_size;
	}

	return ret;
}

int proc_actionfile_write(struct file *file, const char *buffer, unsigned long count, void *data) {
	char temp[NAME_MAX];

	spin_lock(&my_lock);

	action_buffer_size = count;
	if (action_buffer_size > NAME_MAX ) {
		action_buffer_size = NAME_MAX;
	}
	
	if ( copy_from_user(action_buffer, buffer, action_buffer_size) ) {
		return -EFAULT;
	}

	strlcpy(temp, action_buffer, action_buffer_size);

	if (strcmp(temp,"start_collecting") == 0) {
		signalStartCollecting();
	} else if (strcmp(temp,"stop_collecting") == 0) {
		signalStopCollecting();
	} else if (strcmp(temp,"print_stack") == 0) {
		signalPrintStack();
	} else if (strcmp(temp,"reset_stack") == 0) {
		signalResetStack();
	} else if (strcmp(temp,"clear_stack") == 0) {
		signalClearStack();
	} else if (strcmp(temp,"init_stack") == 0) {
		signalInitStack();
	}

	spin_unlock(&my_lock);

	return action_buffer_size;
}


int proc_outputfile_read(char *buffer, char **buffer_location, off_t offset, int buffer_length, int *eof, void *data) {
	int ret;
	
	if (offset > 0) {
		ret  = 0;
	} else {
		memcpy(buffer, output_file_name_buffer, output_file_name_buffer_size);
		ret = output_file_name_buffer_size;
	}

	return ret;
}

int proc_outputfile_write(struct file *file, const char *buffer, unsigned long count, void *data) {
	//char temp[NAME_MAX];

	// get buffer size
	output_file_name_buffer_size = count;
	if (output_file_name_buffer_size > NAME_MAX ) {
		output_file_name_buffer_size = NAME_MAX;
	}
	
	// write data to the buffer 
	if ( copy_from_user(output_file_name_buffer, buffer, output_file_name_buffer_size) ) {
		return -EFAULT;
	}

	signalProgramName(output_file_name_buffer);

	//strlcpy(temp, output_file_name_buffer, output_file_name_buffer_size);

	return output_file_name_buffer_size;
}

int proc_iteration_number_read(char *buffer, char **buffer_location, off_t offset, int buffer_length, int *eof, void *data) {
	int ret;
	
	if (offset > 0) {
		ret  = 0;
	} else {
		memcpy(buffer, iteration_number_buffer, iteration_number_buffer_size);
		ret = iteration_number_buffer_size;
	}

	return ret;
}

int proc_iteration_number_write(struct file *file, const char *buffer, unsigned long count, void *data) {
	// get buffer size
	iteration_number_buffer_size = count;
	if (iteration_number_buffer_size > NAME_MAX ) {
		iteration_number_buffer_size = NAME_MAX;
	}
	
	// write data to the buffer 
	if ( copy_from_user(iteration_number_buffer, buffer, iteration_number_buffer_size) ) {
		return -EFAULT;
	}

	signalIterationNumber(iteration_number_buffer);

	return iteration_number_buffer_size;
}


int proc_num_app_threads_read(char *buffer, char **buffer_location, off_t offset, int buffer_length, int *eof, void *data) {
	int ret;
	
	if (offset > 0) {
		ret  = 0;
	} else {
		memcpy(buffer, num_app_threads_buffer, num_app_threads_buffer_size);
		ret = num_app_threads_buffer_size;
	}

	return ret;
}

int proc_num_app_threads_write(struct file *file, const char *buffer, unsigned long count, void *data) {
	int temp = 0;
	int i;
	char c;

	// get buffer size
	num_app_threads_buffer_size = count;
	if (num_app_threads_buffer_size > NAME_MAX ) {
		num_app_threads_buffer_size = NAME_MAX;
	}
	
	// write data to the buffer 
	if ( copy_from_user(num_app_threads_buffer, buffer, num_app_threads_buffer_size) ) {
		return -EFAULT;
	}

	i = 0;
	do {
		c = num_app_threads_buffer[i++];
		if (c >= '0' && c <= '9'){
			temp = temp * 10 + ( c - '0' );
		}
	} while ( c != '\n' && i < strlen(num_app_threads_buffer) );

	signalNumAppThreads(temp);

	return num_app_threads_buffer_size;
}


int proc_conc_gc_stw_read(char *buffer, char **buffer_location, off_t offset, int buffer_length, int *eof, void *data) {
	int ret;
	
	if (offset > 0) {
		ret  = 0;
	} else {
		memcpy(buffer, conc_gc_stw_buffer, conc_gc_stw_buffer_size);
		ret = conc_gc_stw_buffer_size;
	}

	return ret;
}

int proc_conc_gc_stw_write(struct file *file, const char *buffer, unsigned long count, void *data) {
	char temp[NAME_MAX];

	// get buffer size
	conc_gc_stw_buffer_size = count;
	if (conc_gc_stw_buffer_size > NAME_MAX ) {
		conc_gc_stw_buffer_size = NAME_MAX;
	}
	
	// write data to the buffer 
	if ( copy_from_user(conc_gc_stw_buffer, buffer, conc_gc_stw_buffer_size) ) {
		return -EFAULT;
	}

	strlcpy(temp, conc_gc_stw_buffer, conc_gc_stw_buffer_size);

	if (strcmp(temp,"1") == 0) {
		signalConcGCSTW(true);
	} else if (strcmp(temp,"0") == 0) {
		signalConcGCSTW(false);
	} 

	return conc_gc_stw_buffer_size;
}


static int __init proc_init(void) {
	spin_lock_init(&my_lock);

	proc_file_process_name = create_proc_entry("profiled_process", 0666, NULL);
	
	if (proc_file_process_name == NULL) {
		remove_proc_entry("profiled_process", NULL);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n", "profiled_process");
		return -ENOMEM;
	}

	proc_file_process_name->read_proc  = procfile_read;
	proc_file_process_name->write_proc = procfile_write;
	proc_file_process_name->uid 	   = 0;
	proc_file_process_name->gid 	   = 0;
	proc_file_process_name->size 	   = 0;

	printk(KERN_INFO "/proc/%s created\n", "profiled_process");	

	proc_file_action = create_proc_entry("signal_action", 0666, NULL);
	
	if (proc_file_action == NULL) {
		remove_proc_entry("signal_action", NULL);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n", "signal_action");
		return -ENOMEM;
	}

	proc_file_action->read_proc  = proc_actionfile_read;
	proc_file_action->write_proc = proc_actionfile_write;
	proc_file_action->uid 	     = 0;
	proc_file_action->gid 	     = 0;
	proc_file_action->size 	     = 0;

	printk(KERN_INFO "/proc/%s created\n", "signal_action");	

	proc_output_file_name = create_proc_entry("program_name_output", 0666, NULL);
	
	if (proc_output_file_name == NULL) {
		remove_proc_entry("program_name_output", NULL);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n", "program_name_output");
		return -ENOMEM;
	}

	proc_output_file_name->read_proc  = proc_outputfile_read;
	proc_output_file_name->write_proc = proc_outputfile_write;
	proc_output_file_name->uid 	      = 0;
	proc_output_file_name->gid 	      = 0;
	proc_output_file_name->size 	  = 0;

	printk(KERN_INFO "/proc/%s created\n", "program_name_output");
	
	iteration_number = create_proc_entry("iteration_number", 0666, NULL);
	
	if (iteration_number == NULL) {
		remove_proc_entry("iteration_number", NULL);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n", "iterattion_number");
		return -ENOMEM;
	}

	iteration_number->read_proc  = proc_iteration_number_read;
	iteration_number->write_proc = proc_iteration_number_write;
	iteration_number->uid 	     = 0;
	iteration_number->gid 	     = 0;
	iteration_number->size 	     = 0;

	printk(KERN_INFO "/proc/%s created\n", "iteration_number");


	proc_num_app_threads = create_proc_entry("num_app_threads", 0666, NULL);
	
	if (proc_num_app_threads == NULL) {
		remove_proc_entry("num_app_threads", NULL);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n", "num_app_threads");
		return -ENOMEM;
	}

	proc_num_app_threads->read_proc  = proc_num_app_threads_read;
	proc_num_app_threads->write_proc = proc_num_app_threads_write;
	proc_num_app_threads->uid 	     = 0;
	proc_num_app_threads->gid 	     = 0;
	proc_num_app_threads->size 	     = 0;

	printk(KERN_INFO "/proc/%s created\n", "num_app_threads");


	proc_conc_gc_stw = create_proc_entry("conc_gc_stw", 0666, NULL);
	
	if (proc_conc_gc_stw == NULL) {
		remove_proc_entry("conc_gc_stw", NULL);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n", "conc_gc_stw");
		return -ENOMEM;
	}

	proc_conc_gc_stw->read_proc  = proc_conc_gc_stw_read;
	proc_conc_gc_stw->write_proc = proc_conc_gc_stw_write;
	proc_conc_gc_stw->uid 	     = 0;
	proc_conc_gc_stw->gid 	     = 0;
	proc_conc_gc_stw->size 	     = 0;

	printk(KERN_INFO "/proc/%s created\n", "conc_gc_stw");

	return 0;	/* everything is ok */
}

static void __exit proc_exit(void) {
	remove_proc_entry("profiled_process", NULL);
	printk(KERN_INFO "/proc/%s removed\n", "profiled_process");

	remove_proc_entry("signal_action", NULL);
	printk(KERN_INFO "/proc/%s removed\n", "signal_action");

	remove_proc_entry("program_name_output", NULL);
	printk(KERN_INFO "/proc/%s removed\n", "program_name_output");

	remove_proc_entry("iteration_number", NULL);
	printk(KERN_INFO "/proc/%s removed\n", "iteration_number");

	remove_proc_entry("num_app_threads", NULL);
	printk(KERN_INFO "/proc/%s removed\n", "num_app_threads");

	remove_proc_entry("conc_gc_stw", NULL);
	printk(KERN_INFO "/proc/%s removed\n", "conc_gc_stw");
}

module_init(proc_init);
module_exit(proc_exit);

MODULE_LICENSE("GPL"); 
